;ATTiny24 with external 32 kHz (32768 Hz) low-frequency crystal
;Fuses: low = 0xC6, high = 0xDD, ext = 0xFF

;simple quartz clock for 4 7-segment LED displays (multiplexed)
;(C) Arne Rossius 2008-02-21, http://www.ebps.de.vu/

;PORTS:
;  PA0 = "Minutes" key (to ground)
;  PA1 = "Hours" key (to ground)
;  PA2 = Colon (active low)
;  PA3 = Shift Register Strobe
;  PA4 = Shift Register Clock
;  PA5 = Shift Register Data (active low, g/f/a/b/e/d/c/dp, g first, Hour 10's first)
;  PA6 = [unused]
;  PA7 = [unused]
;  PB0 = XTAL
;  PB1 = XTAL
;  PB2 = Shift Register Output Enable (PWM for dimming)

.equ	DOT_FREQ = 7 ;flashing dot: 0->1Hz, 1->.5Hz, 2->.25Hz, 7->static on
.equ	KEY_REPEATSTART = 16 ;16/16s = 1s until repeating starts
.equ	KEY_REPEAT = 5 ;16/5 = 3.2 repeats per second

.include "../include/tn24def.inc"
 
.def	temp = R16
.def	temp2 = R17
.def	min1 = R18
.def	min10 = R19
.def	hr1 = R20
.def	hr10 = R21
.def	sreg_data0 = R22
.def	sreg_data1 = R23
.def	repeat = R24

;R31:R30: Z pointer for Flash access

.equ	Min = 0
.equ	Hr = 1
.equ	Colon = 2
.equ	Strobe = 3
.equ	Clock = 4
.equ	Data = 5

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org PCI0addr
	rjmp	keytoggle
.org OVF0addr
	rjmp	keyrepeat

;===============================================================================

disptab: ;MSB = g / f / a / b / e / d / c / dp = LSB
	.db 0x7e, 0x12, 0xbc, 0xb6, 0xd2, 0xe6, 0xee, 0x32
	.db 0xfe, 0xf6, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

;===============================================================================

keytoggle:
	;pin change interrupt for hour/minute keys
	push	temp
	in	temp, SREG
	push	temp
	
	in	temp, PINA
	andi	temp, 0x03
	cpi	temp, 0x03
	breq	key_released
	
	;key pressed: reset and enable debounce/repeat timer
	ldi	repeat, 0
	ldi	temp, 0
	out	TCNT0, temp
	ldi	temp, 0x01 ;Clk/1 => ~8ms debouncing time
	out	TCCR0B, temp
	
keytoggle_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_released:
	;key released: disable debounce/repeat timer
	ldi	temp, 0x00
	out	TCCR0B, temp
	rjmp	keytoggle_end

;===============================================================================

keyrepeat:
	;key debounce/repeat timer overflow
	push	temp
	in	temp, SREG
	push	temp
	
	;change from debounce to repeat mode
	cpi	repeat, 0
	brne	repeatdiv
	ldi	temp, 0x02 ;Clk/8
	out	TCCR0B, temp
	ldi	temp, 0
	out	TCNT0, temp
	ldi	repeat, KEY_REPEATSTART
	rjmp	key_action
	
repeatdiv:
	;repeated keypress
	dec	repeat
	brne	keyrepeat_end
	ldi	repeat, KEY_REPEAT
key_action:
	;preform key action
	sbis	PINA, Hr
	rjmp	key_hr
	sbis	PINA, Min
	rjmp	key_min
	
keyrepeat_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_hr:
	;"Hour" key: increment hours
	rcall	inc_hr
	rcall	update_disp
	rjmp	keyrepeat_end

key_min:
	;"Minute" key: increment minutes and reset seconds
	rcall	inc_min
	rcall	update_disp
	ldi	temp, 0
	out	TCNT1H, temp
	out	TCNT1L, temp
	rjmp	keyrepeat_end
	
;===============================================================================

reset:
	;stack pointer setup
	ldi	temp, RAMEND
	out	SPL, temp
	
	;port setup
	ldi	temp, 0xFC
	out	DDRA, temp
	ldi	temp, 0x03
	out	PORTA, temp
	ldi	temp, 0x04
	out	DDRB, temp
	ldi	temp, 0x04
	out	PORTB, temp
	
	;Timer 0 setup (Key debounce/repeat timer)
	ldi	temp, 1<<TOIE0
	out	TIMSK0, temp
	
	;Timer 1 setup (Clock timer)
	ldi	temp, 1<<WGM12 | 0x03 ;CTC Mode, Clk/64
	out	TCCR1B, temp
	ldi	temp, HIGH(30720) ;overflow every 60s (32768/(64*30720) = 1/60)
	out	OCR1AH, temp
	ldi	temp, LOW(30720)
	out	OCR1AL, temp
	
	;Pin Change interrupt setup
	ldi	temp, 0x03
	out	PCMSK0, temp
	ldi	temp, 1<<PCIE0
	out	GIMSK, temp
	
	;register setup
	ldi	hr10, 0
	ldi	hr1, 0
	ldi	min10, 0
	ldi	min1, 0
	rcall	update_disp
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------
	
loop:
	;increment clock if timer reached 60s
	sbis	TIFR1, OCF1A
	rjmp	timeinc_end
	ldi	temp, 1<<OCF1A ;reset flag
	out	TIFR1, temp
	rcall	inc_min
	rcall	update_disp
timeinc_end:

	;flashing colon
	ldi	temp, 0x03
	in	temp2, TCNT1L ;need to read TCNT1L before TCNT1H can be read
	in	temp2, TCNT1H
	sbrc	temp2, DOT_FREQ
	ori	temp, 1<<Colon
	out	PORTA, temp
	
	rjmp	loop
	
;===============================================================================

inc_min:
	;increment minutes
	inc	min1
	cpi	min1, 10
	brlo	inc_end
	clr	min1
	inc	min10
	cpi	min10, 6
	brlo	inc_end
	clr	min10
inc_hr:
	;increment hours
	inc	hr1
	cpi	hr1, 10
	brlo	hr_24
	clr	hr1
	inc	hr10
	rjmp	inc_end
hr_24:
	cpi	hr1, 4
	brne	inc_end
	cpi	hr10, 2
	brne	inc_end
	clr	hr1
	clr	hr10
inc_end:
	ret
	
;===============================================================================

update_disp:
	;initialize registers for shifting
	in	sreg_data0, PORTA
	andi	sreg_data0, ~(1<<Data | 1<<Clock)
	in	sreg_data1, PORTA
	andi	sreg_data1, ~(1<<Data | 1<<Clock)
	ori	sreg_data1, 1<<Data
	
	;convert time to 7-segment display representation and shift out
	ldi	ZH, 0
	ldi	ZL, LOW(disptab*2)
	add	ZL, hr10
	lpm	temp, Z
	rcall	shift
	ldi	ZL, LOW(disptab*2)
	add	ZL, hr1
	lpm	temp, Z
	rcall	shift
	ldi	ZL, LOW(disptab*2)
	add	ZL, min10
	lpm	temp, Z
	rcall	shift
	ldi	ZL, LOW(disptab*2)
	add	ZL, min1
	lpm	temp, Z
	rcall	shift
	
	;latch new data
	sbi	PORTA, Strobe
	cbi	PORTA, Strobe
	
	ret
	
;--------------------

shift:
	;shift out 8 bits (inverted)
	out	PORTA, sreg_data0 ;Bit 7
	sbrs	temp, 7
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 6
	sbrs	temp, 6
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 5
	sbrs	temp, 5
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 4
	sbrs	temp, 4
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 3
	sbrs	temp, 3
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 2
	sbrs	temp, 2
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 1
	sbrs	temp, 1
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 0
	sbrs	temp, 0
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	ret
	